@@ -158,7 +158,7 @@ def get_self_bind_sock() -> socket:
 class SecureReliableSocket(socket):
     __slots__ = [
         "timeout", "span", "address", "shared_key", "mtu_size",
-        "sender_seq", "sender_buffer", "sender_signal", "sender_buffer_lock",
+        "sender_seq", "sender_buffer", "sender_signal", "sender_buffer_lock", "sender_socket",
         "receiver_seq", "receiver_unread_size", "receiver_buffer",
         "broadcast_hook_fnc", "loss", "try_connect", "established"]
 
@@ -180,6 +180,7 @@ class SecureReliableSocket(socket):
         self.sender_buffer: Deque[Packet] = deque()
         self.sender_signal = threading.Event()  # clear when buffer is empty
         self.sender_buffer_lock = threading.Lock()
+        self.sender_socket: Optional[socket] = None
         # receiver params
         self.receiver_seq = CycInt(1)  # next receive sequence
         self.receiver_unread_size = 0
@@ -192,7 +193,7 @@ class SecureReliableSocket(socket):
         self.try_connect = False
         self.established = False
 
-    def connect(self, address: _WildAddress) -> None:
+    def connect(self, address: _WildAddress, listen_port: int = None) -> None:
         """
         `connection()` establishment process
         ====
@@ -237,22 +238,33 @@ class SecureReliableSocket(socket):
 
         # bind socket address
         self.address = address = get_formal_address_format(address, self.family)
-        address_copy = list(address)
-        address_copy[0] = ""  # the port is CLOSE_WAIT state if this raise OSError
-        self.bind(tuple(address_copy))
+        bind_addr = list(address)
+
+        if listen_port is None:
+            # global
+            bind_addr[0] = ""
+        else:
+            # local (for debug)
+            bind_addr[0] = "localhost"
+            bind_addr[1] = listen_port
+            self.sender_socket = socket(self.family, s.SOCK_DGRAM)
+
+        # the port is CLOSE_WAIT state if this raise OSError
+        self.bind(tuple(bind_addr))
         log.debug("try to communicate with {}".format(address))
 
         # warning: allow only 256bit curve
         select_curve = ecdsa.curves.NIST256p
         log.debug("select curve {} (static)".format(select_curve))
+        assert select_curve.baselen == 32, ("curve is 256bits size only", select_curve)
 
         # 1. UDP hole punching
         punch_msg = b"udp hole punching"
         self.sendto(S_HOLE_PUNCHING + punch_msg + select_curve.name.encode(), address)
 
         # my secret & public key
-        my_sk: Optional[ecdsa.SigningKey] = None
-        my_pk: Optional[ecdsa.VerifyingKey] = None
+        my_sk = ecdsa.SigningKey.generate(select_curve)
+        my_pk = my_sk.get_verifying_key()
 
         # other's public key
         other_pk: Optional[ecdsa.VerifyingKey] = None
@@ -267,19 +279,15 @@ class SecureReliableSocket(socket):
                 if stage == S_HOLE_PUNCHING:
                     # 2. send my public key
                     curve_name = data.replace(punch_msg, b'').decode()
+                    other_curve = find_ecdhe_curve(curve_name)
+                    assert select_curve == other_curve, ("different curve", select_curve, other_curve)
                     select_curve = find_ecdhe_curve(curve_name)
-                    # update curve
-                    my_sk = ecdsa.SigningKey.generate(select_curve)
-                    my_pk = my_sk.get_verifying_key()
                     self.sendto(S_SEND_PUBLIC_KEY + my_pk.to_string(), address)
                     log.debug("success UDP hole punching")
 
                 elif stage == S_SEND_PUBLIC_KEY:
                     # 3. get public key & send shared key
                     other_pk = ecdsa.VerifyingKey.from_string(data, select_curve)
-                    # using my select curve
-                    my_sk = ecdsa.SigningKey.generate(select_curve)
-                    my_pk = my_sk.get_verifying_key()
                     shared_point = my_sk.privkey.secret_multiplier * other_pk.pubkey.point
                     self.shared_key = sha256(shared_point.x().to_bytes(32, 'big')).digest()
                     shared_key = os.urandom(32)
@@ -333,7 +341,7 @@ class SecureReliableSocket(socket):
         # avoid Path MTU Discovery Blackhole
         if self.family == s.AF_INET:
             self.setsockopt(s.IPPROTO_IP, IP_MTU_DISCOVER, IP_PMTUDISC_DO)
-        self.mtu_size = self._find_mut_size()
+        self.mtu_size = self._find_mtu_size()
         if self.family == s.AF_INET:
             self.setsockopt(s.IPPROTO_IP, IP_MTU_DISCOVER, IP_PMTUDISC_DONT)
         log.debug("success get MUT size %db", self.mtu_size)
@@ -345,7 +353,7 @@ class SecureReliableSocket(socket):
         # auto exit when program closed
         atexit.register(self.close)
 
-    def _find_mut_size(self) -> int:
+    def _find_mtu_size(self) -> int:
         """confirm by submit real packet"""
         wait = 0.05
         mut = 1472  # max ipv4:1472b, ipv6:1452b
@@ -429,10 +437,6 @@ class SecureReliableSocket(socket):
                 data, _addr = self.recvfrom(65536)
                 packet = bin2packet(self._decrypt(data))
 
-                # reject too early or late packet
-                if 3600.0 < abs(time() - packet.time):
-                    continue
-
                 last_receive_time = time()
                 # log.debug("r<< %s", packet)
             except ValueError:
@@ -449,11 +453,13 @@ class SecureReliableSocket(socket):
                 with self.sender_buffer_lock:
                     if 0 < len(self.sender_buffer):
                         for i in range(packet.sequence - self.sender_buffer[0].sequence + 1):
+                            # remove packet that sent and confirmed by ACK
                             self.sender_buffer.popleft()
                             if len(self.sender_buffer) == 0:
                                 break
                         if not self._send_buffer_is_full():
                             self.sender_signal.set()
+                            log.debug("allow sending operation again seq={}".format(packet.sequence))
                 continue
 
             # receive reset
@@ -558,7 +564,7 @@ class SecureReliableSocket(socket):
             # reached EOF & push broadcast packets
             if packet.control & CONTROL_EOF:
                 # note: stopped sending broadcast packet after main stream for realtime
-                log.debug("reached EOF of data")
+                log.debug("reached end of chunk seq={}".format(packet.sequence))
 
             # update last packet
             last_packet = packet
@@ -579,6 +585,7 @@ class SecureReliableSocket(socket):
 
     def get_window_size(self) -> int:
         """maximum size of data you can send at once"""
+        # Packet = [nonce 16b][tag 16b][static 14b][data xb]
         return self.mtu_size - 32 - packet_struct.size
 
     def send(self, data: bytes, flags: int = 0) -> int:
@@ -591,17 +598,20 @@ class SecureReliableSocket(socket):
         """warning: row-level method"""
         if not self.established:
             raise ConnectionAbortedError('disconnected')
-        # decrease 1byte for padding of AES when packet is full size
+
         window_size = self.get_window_size()
         length = len(data) // window_size
         send_size = 0
         for i in range(length + 1):
-            control = 0
+            # control flag
+            control = 0b00000000
             buffer_is_full = self._send_buffer_is_full()
             if i == length or buffer_is_full:
                 control |= CONTROL_PSH
             if i == length:
                 control |= CONTROL_EOF
+
+            # send one packet
             throw = data[window_size * i:window_size * (i + 1)]
             with self.sender_buffer_lock:
                 packet = Packet(control, self.sender_seq, 0, time(), throw.tobytes())
@@ -609,25 +619,40 @@ class SecureReliableSocket(socket):
                 self.sendto(self._encrypt(packet2bin(packet)), self.address)
                 self.sender_seq += 1
             send_size += len(throw)
+
+            # note: sleep a mSec to avoid packet loss
             if window_size == len(throw):
-                # warning: need wait when send full size chunk
                 sleep(FULL_SIZE_PACKET_WAIT)
+
             # block sendall() when buffer is full
             if buffer_is_full:
                 self.sender_signal.clear()
+                log.debug("buffer is full and wait for signaled")
                 break
         return send_size
 
+    def sendto(self, data: bytes, address: _Address) -> int:  # type: ignore
+        """note: sendto() after bind() with different port cause OSError on recvfrom()"""
+        if self.is_closed:
+            return 0
+        elif self.sender_socket is None:
+            return super().sendto(data, address)
+        else:
+            return self.sender_socket.sendto(data, address)
+
     def sendall(self, data: bytes, flags: int = 0) -> None:
         """high-level method, use this instead of send()"""
         assert flags == 0, "unrecognized flags"
-        if not self._send_buffer_is_full():
-            self.sender_signal.set()
         send_size = 0
         data = memoryview(data)
         while send_size < len(data):
+            if not self._send_buffer_is_full():
+                self.sender_signal.set()
             if self.sender_signal.wait(self.timeout):
                 send_size += self._send(data[send_size:])
+            else:
+                log.debug("waiting for sending buffer have space..")
+        log.debug("send operation success %sb", send_size)
 
     def broadcast(self, data: bytes) -> None:
         """broadcast data (do not check reach)"""
@@ -721,6 +746,8 @@ class SecureReliableSocket(socket):
             sleep(0.001)
             super().close()
             self.receiver_buffer.close()
+            if self.sender_socket is not None:
+                self.sender_socket.close()
             atexit.unregister(self.close)
 
 
